/*-------------------------------------------------------------------
 *
 *  @file blt_server.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include "blt_server.h"
#include <getopt.h>


struct mcp_t *master;	/**< holds data for master control thread */
int con_run;				/**< console status */

/**
 *  @brief handle all kind of signals!
 *
 *  @param signal number
 *
 *  @returns nothing
 */
void sigint_handler(int signum)
{
	if (master != NULL) {
		/* 
		 *  don't exit on HUP, just reload/save databases
		 */
		if (signum == SIGHUP) {
			blt_ln_load(&master->server.locationNames);
			blt_dn_save(&master->server.devNames);
		}
		else {
			if (master->running == 1) {
				printf("received signal...\ndumping devices names...\nexiting!\n");
			}
			master->running = 0;
			con_run = 0;
		}
	}
}

/**
 *  @brief thread that controls the complete server
 *
 *  @param args pointer to mcp_t
 *
 *  @returns nothing
 */
void* master_control_program(void *args)
{
	struct mcp_t *myMaster = (struct mcp_t*) args;
	int ti, tc;
	

	/* set last inquiry to current time */
	myMaster->lastInq = time(NULL);
	
	/* main loop */
	while (myMaster->running) {
		
		/* is it time for an INQUIRY ? */
		myMaster->curInq = time(NULL);
		
		if (myMaster->curInq >= myMaster->lastInq + myMaster->inqStep) {
			myMaster->lastInq = myMaster->curInq;
			
			blt_dn_save(&myMaster->server.devNames);
			
			ti = 0; tc = 0;
			do {
				if (myMaster->server.trackers[ti].myListIndex != -1) {
					myMaster->server.trackers[ti].doInq = 1;
					pthread_cond_signal(&myMaster->server.trackers[ti].actionSig);
					tc++;
				}
				ti++;
			} while (ti < MAX_NUM_TRACKERS && tc < myMaster->server.numTrackers);
		}

		/* wakeup thread to check for DEVICE_NAMES */
		sleep(myMaster->wakeupDelay);
		ti = 0; tc = 0;
		
		#ifdef DEBUG
		fprintf(stderr, "ts: sending wakeup to all clients\n");
		#endif
		
		do {
			if (myMaster->server.trackers[ti].myListIndex != -1) {
				if (myMaster->server.trackers[ti].status == BLT_TC_SLEEPING) {
					myMaster->server.trackers[ti].status = BLT_TC_READ;
					pthread_cond_signal(&myMaster->server.trackers[ti].actionSig);
				}
				tc++;
			}
			ti++;
		} while (ti < MAX_NUM_TRACKERS && tc < myMaster->server.numTrackers);
		
	} /* while master->running == 1*/

	pthread_exit((void*)NULL);
}

/**
 *  @brief main program of blt_server
 *
 *  @param argc number of arguments
 *  @param argv argument vector
 *
 *  @returns 1 on success or 0 on failure
 */
int main(int argc, char **argv)
{
	int cmd_size = 0;
	char *cmd_line = NULL;
	pthread_t mcp_thread;
	char c;
	char *p;


	/* alloc global memory */
	master = malloc(sizeof(struct mcp_t));
	memset(master, 0, sizeof(struct mcp_t));

	/* use console */
	con_run = 1;
	
	/* get arguments */
	while ((c = getopt(argc, argv, "s:d:i:p:n:I:vc")) != -1) {
		switch (c) {
			case 's':
				master->server.tracker_ip = (char*) malloc(strlen(optarg)+1);
				strcpyendchr(optarg, master->server.tracker_ip, ':', strlen(optarg));
				p = strchr(optarg, ':');
				if (p > 0) master->server.tracker_port = atoi(p+1);
				break;
			case 'd':
				master->server.data_ip = (char*) malloc(strlen(optarg)+1);
				strcpyendchr(optarg, master->server.data_ip, ':', strlen(optarg));
				p = strchr(optarg, ':');
				if (p > 0) master->server.data_port = atoi(p+1);
				break;
			case 'i':
				master->server.defaultCfg.inquiryDuration = atoi(optarg);
				break;
			case 'p':
				master->server.defaultCfg.pageTimeout = atoi(optarg);
				break;
			case 'n':
				master->server.defaultCfg.maxAskName = atoi(optarg);
				break;
			case 'I':
				master->inqStep = atoi(optarg);
				break;
			case 'w':
				master->wakeupDelay = atoi(optarg);
				break;
			case 'c':
				con_run = 2;
				break;
			case '?':
				fprintf(stderr,	"syntax: %s -vsdipnIwc\n\n"\
										"\t-s<host:port>\tIP and port to listen for tracker clients\n"\
										"\t\t\tdefault is 0.0.0.0:4444\n"\
										"\t-d<host:port>\tIP and port to listen for data clients\n"\
										"\t\t\tdefault is 0.0.0.0:4445\n"\
										"\t-i<slots>\tinquiry duration in 1.28s units (slots)\n"\
										"\t\t\tdefault is 20\n"\
										"\t-p<slots>\tpage timeout in 1.28s units (slots)\n"\
										"\t\t\tdefault is 10\n"\
										"\t-n<count>\tmaximum number of name request per inquiry cycle\n"\
										"\t\t\tdefault is 10\n"\
										"\t-I<seconds>\tdelay between inquirys (in seconds)\n"\
										"\t\t\tdefault is 300 (5min)\n"\
										"\t-w<seconds>\twakeup delay for tracker clients (only for fine tuning)\n"\
										"\t\t\tdefault is 20 seconds\n"\
										"\n"\
										"\t-c\t\tnon interactive mode (no console!)\n"\
										"\t\t\tdefault is console active\n"\
										"\t-v\t\tshow version string\n"\
										"", argv[0]);
				return(0);
			case 'v':
				printf(BLT_SERVER_BANNER);
				return(1);
				break;
			default:
				break;
		}
	}
		
	/* tracker */
	if (master->server.tracker_ip == NULL) {
		master->server.tracker_ip = (char*) malloc(8);
		memset(master->server.tracker_ip, 0, 8);
		memcpy(master->server.tracker_ip, "0.0.0.0", 7);
	}
	if (master->server.tracker_port == 0) {
		master->server.tracker_port = 4444;
	}
	master->server.numTrackers = 0;

	/* data */
	if (master->server.data_ip == NULL) {
		master->server.data_ip = (char*) malloc(8);
		memset(master->server.data_ip, 0, 8);
		memcpy(master->server.data_ip, "0.0.0.0", 7);
	}
	if (master->server.data_port == 0) {
		master->server.data_port = 4445;
	}
	master->server.numDatas = 0;

	/* init default tracker config. */
	master->server.defaultCfg.pageScan = 1;
	master->server.defaultCfg.inquiryScan = 1;
	master->server.defaultCfg.maxNumDevices = 0;
	master->server.defaultCfg.name = NULL; /* use default name */
	
	if (master->server.defaultCfg.inquiryDuration == 0)
		master->server.defaultCfg.inquiryDuration = 20;
	if (master->server.defaultCfg.pageTimeout == 0)
		master->server.defaultCfg.pageTimeout = 10;
	if (master->server.defaultCfg.maxAskName == 0)
		master->server.defaultCfg.maxAskName = 10;
		
	if (master->inqStep == 0) master->inqStep = 300;
	if (master->wakeupDelay == 0) master->wakeupDelay = 20;	


	#ifdef DEBUG4
	printf("id: %d, pt: %d, ma: %d, is: %d\n",master->server.defaultCfg.inquiryDuration,
	master->server.defaultCfg.pageTimeout,master->server.defaultCfg.maxAskName,master->inqStep);
	/*sleep(5);*/
	#endif

	/* init SSL */
	SSL_load_error_strings();
	SSL_library_init();       
	RAND_load_file("/dev/urandom", 4096);
	
	/* init location names and login/pw */
	blt_ln_init(&master->server.locationNames);
	blt_ln_load(&master->server.locationNames);

	/* init device store */
	blt_dn_init(&master->server.devNames);
	/* load known devices from last session */
	blt_dn_load(&master->server.devNames);
	
	/* init the tracker and the data threads */
	trackers_init(&master->server);
	datas_init(&master->server);
	
#ifdef BLT_SSL
	/* SSL ctx setup */
	master->server.ssl_ctx = SSL_CTX_new(SSLv23_server_method());
	if (SSL_CTX_use_RSAPrivateKey_file(master->server.ssl_ctx, CERT_FILEPATH, SSL_FILETYPE_PEM) != 1) {
		fprintf(stderr, "SSL cert.pem failed\n");
		exit(1);
	}
	if (SSL_CTX_use_certificate_chain_file(master->server.ssl_ctx, CERT_FILEPATH) != 1) {
		fprintf(stderr, "SSL chain failed\n");
		exit(1);
	}
	SSL_CTX_set_cipher_list(master->server.ssl_ctx, "DEFAULT");
#endif

	/* init and start main push thread */
	blt_push_init(&master->server.pushServer);
	pthread_create(&master->push_serv, NULL, blt_pusher, (void*) &master->server);
	
	/* start the two network threads */
	pthread_create(&master->tc_serv, NULL, trackers_wait_for_clients, (void*) &master->server);
	pthread_create(&master->dc_serv, NULL, datas_wait_for_clients, (void*) &master->server);	
	
	/* signalling */
	signal(SIGINT, sigint_handler); /* Ctrl-C */
	signal(SIGHUP, sigint_handler); /* kill -HUP */
	signal(SIGTERM, sigint_handler); /* kill */
	signal(SIGPIPE, SIG_IGN); /* this keeps us from crashing while send data */

	/* print banner to stdout */
	printf(BLT_SERVER_BANNER);

	master->running = 1;
	/* start master control thread */
	pthread_create(&mcp_thread, NULL, master_control_program, (void*)master);

	/* this is the server console */
	do {
		if (con_run == 1) { /* interactive mode */
			printf("BLT=>");
			if (getline(&cmd_line, &cmd_size, stdin) > 0) {

				if (strncasecmp(cmd_line, "quit", 4) == 0) {
					printf("BLT:QUIT\n");
					master->running = 0;
					con_run = 0;
				}
			}
			if (cmd_line != NULL) {
				free(cmd_line);
				cmd_size = 0;
			}
		}
		else { /* non interactive mode */
			sleep(300);
		}
	} while(con_run != 0);

	
	/* save known devices for next session */
	blt_dn_save(&master->server.devNames);
	
	return(1);
}
