/*-------------------------------------------------------------------
 *
 *  @file data.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include "blt_server.h"

/**
 *  @brief print device status to fd
 *
 *  @param dev pointer to device status to print
 *  @param fp  output file descriptor
 *
 *  @returns 1 on success or 0 on failure
 */
int blt_dc_printDev(struct blt_device_status_t *dev, int fd)
{
	char ts[30];
	struct tm tmp_time;
	char buffer[1024];
	char buffer2[1024];


	assert(dev != NULL);

	memset(&tmp_time, 0, sizeof(struct tm));
	localtime_r(&dev->tStamp, &tmp_time);
	strftime(ts, 21, "%m.%d.%Y %T", &tmp_time);

	/* escape bad characters */
	memset(buffer2, 0, 1024);
	printWithEsc(dev->dev.name, buffer2, 1024);

	/* format: bdaddr, "name", class, rssi, "date_time", latitude, longitude */
	memset(buffer, 0, 1024);
	sprintf(buffer, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X, \"%s\", 0x%.6X, 0x%.8X, \"%s\", %d, %d\r\n",
		dev->dev.bdAddr[0], dev->dev.bdAddr[1], dev->dev.bdAddr[2],
		dev->dev.bdAddr[3], dev->dev.bdAddr[4], dev->dev.bdAddr[5],
		buffer2, dev->dev.devClass, dev->rssi, ts, dev->la, dev->lg);
		
	if (nwrite(fd, buffer, strlen(buffer)) == -1) return(0);
	else return(1);
}

/**
 *  @brief data client thread
 *
 *  @param args pointer to dataclient session
 *
 *  @returns nothing 
 */
void *blt_data_client(void *args)
{
	struct blt_data_client_session_t *self; /* = (struct blt_data_client_session_t*) args; */ /**< self reference */
	char buffer[1024];	/**< receive buffer */
	int rb = 0;				/**< received bytes */
	int i, i2, p;			/**< loop counter */
	struct blt_device_status_t *dev;
	struct blt_device_status_t filter;
	unsigned char useFilter = 0;
	struct blt_push_t *event;
	char *ps;
	struct readbuf_t buf;
	
	
	assert(args != NULL);
	self = (struct blt_data_client_session_t*) args;
	
	/* init readbuf lib */
	readbufInit(&buf, 4096, NULL);
	
	#ifdef DEBUG
	fprintf(stderr, "dc[%d] ready\n", self->myListIndex);
	#endif
		
	/* --- getting --- */	
	do { /* main "get" loop */
		if ((rb = readbuf(&buf, self->mySocket, buffer, 1024, 300)) == 1) {
		
			/* -- get stuff -- */
			if (strstr(buffer, "data(num_total_devices)") == buffer) {
				i = 0; p = 0; i2 = 0;
				do {
					if (self->server->trackers[i].myListIndex != -1) {
						pthread_rwlock_rdlock(&self->server->trackers[i].pubDevicesLock);
						i2 += self->server->trackers[i].numPubDevices;
						pthread_rwlock_unlock(&self->server->trackers[i].pubDevicesLock);
						p++;
					}
					i++;
				} while (p < self->server->numTrackers && i < MAX_NUM_TRACKERS);
				sprintf(buffer, "%d\r\n",i2);
				if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
					self->status = BLT_DC_DISCONNECTED;
				}
			}
			if (strstr(buffer, "data(num_devices_at") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				sscanf(buffer, "data(num_devices_at, %d, %d)", &filter.la, &filter.lg);
				i = 0; p = 0; i2 = 0;
				do {
					if (self->server->trackers[i].myListIndex != -1) {
						
						pthread_rwlock_rdlock(&self->server->trackers[i].pubDevicesLock);
						dev = self->server->trackers[i].pubDevices;
						if (self->server->trackers[i].numPubDevices > 0) {
							if (filter.la == dev[0].la || filter.lg == dev[0].lg) {
							
								sprintf(buffer, "%d\r\n", self->server->trackers[i].numPubDevices);

								if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
									self->status = BLT_DC_DISCONNECTED;
									p = MAX_NUM_TRACKERS;
								}
								
								/* stop searching */
								p = MAX_NUM_TRACKERS;
								i2 = 1;
							}
						}
						pthread_rwlock_unlock(&self->server->trackers[i].pubDevicesLock);
						p++;
					}
					i++;
				} while (p < self->server->numTrackers && i < MAX_NUM_TRACKERS);
				/* default is null (0) */
				if (i2 == 0) {
					sprintf(buffer, "0\r\n");
					if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
						self->status = BLT_DC_DISCONNECTED;
					}
				}
			}
			else if (strstr(buffer, "data(list_all_devices)") == buffer) {
				i = 0; p = 0;
				do {
					if (self->server->trackers[i].myListIndex != -1) {
						
						pthread_rwlock_rdlock(&self->server->trackers[i].pubDevicesLock);						
						dev = self->server->trackers[i].pubDevices;
						for (i2 = 0; i2 < self->server->trackers[i].numPubDevices; i2++) {
							if (!blt_dc_printDev(&dev[i2], self->mySocket)) {
								self->status = BLT_DC_DISCONNECTED;
								p = MAX_NUM_TRACKERS;
								break;
							}
						}
						pthread_rwlock_unlock(&self->server->trackers[i].pubDevicesLock);
						p++;
					}
					i++;
				} while (p < self->server->numTrackers && i < MAX_NUM_TRACKERS);
				if (self->status != BLT_DC_DISCONNECTED) {
					sprintf(buffer, "data(list_complete)\r\n");
					if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
						self->status = BLT_DC_DISCONNECTED;
					}
				}
			}
			else if (strstr(buffer, "data(list_devices_at") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				sscanf(buffer, "data(list_devices_at, %d, %d)", &filter.la, &filter.lg);
				i = 0; p = 0;
				do {
					if (self->server->trackers[i].myListIndex != -1) {
						pthread_rwlock_rdlock(&self->server->trackers[i].pubDevicesLock);						
						dev = self->server->trackers[i].pubDevices;
						assert(dev != NULL);
						for (i2 = 0; i2 < self->server->trackers[i].numPubDevices; i2++) {
							if (filter.la == dev[i2].la || filter.lg == dev[i2].lg) {
								if (!blt_dc_printDev(&dev[i2], self->mySocket)) {
									self->status = BLT_DC_DISCONNECTED;
									p = MAX_NUM_TRACKERS;
									break;
								}
								
								/*
								 * stop looping thru all trackers  
								 * if we did find the right one
								 *
								 */
								p = MAX_NUM_TRACKERS;
							}
							else {
								/* switch to next tracker */
								break;
							}
						}
						pthread_rwlock_unlock(&self->server->trackers[i].pubDevicesLock);
						p++;
					}
					i++;
				} while (p < self->server->numTrackers && i < MAX_NUM_TRACKERS);
				if (self->status != BLT_DC_DISCONNECTED) {
					sprintf(buffer, "data(list_complete)\r\n");
					if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
						self->status = BLT_DC_DISCONNECTED;
					}
				}
			}
			else if (strstr(buffer, "data(list_device") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				ps = strstr(buffer, ", ");
				if (ps > 0) parsemac(ps+2, filter.dev.bdAddr);
				else continue;
				i = 0; p = 0;
				do {
					if (self->server->trackers[i].myListIndex != -1) {
						pthread_rwlock_rdlock(&self->server->trackers[i].pubDevicesLock);						
						dev = self->server->trackers[i].pubDevices;
						assert(dev != NULL);
						for (i2 = 0; i2 < self->server->trackers[i].numPubDevices; i2++) {
							if (memcmp(dev[i].dev.bdAddr, filter.dev.bdAddr, 6) == 0) {
								if (!blt_dc_printDev(&dev[i2], self->mySocket)) {
									self->status = BLT_DC_DISCONNECTED;
									p = MAX_NUM_TRACKERS;
									break;
								}
							}
						}
						pthread_rwlock_unlock(&self->server->trackers[i].pubDevicesLock);
						p++;
					}
					i++;
				} while (p < self->server->numTrackers && i < MAX_NUM_TRACKERS);
				if (self->status != BLT_DC_DISCONNECTED) {
					sprintf(buffer, "data(list_complete)\r\n");
					if (!nwrite(self->mySocket, buffer, strlen(buffer))) {
						self->status = BLT_DC_DISCONNECTED;
					}
				}
			}
			/* -- push stuff -- */
			else if (strstr(buffer, "data(push_all_devices)") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				useFilter = 0;
				self->status = BLT_DC_PUSH;
			}
			else if (strstr(buffer, "data(push_devices_at") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				sscanf(buffer, "data(push_devices_at, %d, %d)", &filter.la, &filter.lg);
				useFilter = 1;
				self->status = BLT_DC_PUSH;
			}
			else if (strstr(buffer, "data(push_device") == buffer) {
				memset(&filter, 0, sizeof(struct blt_device_status_t));
				ps = strstr(buffer, ", ");
				if (ps > 0) parsemac(ps+2, filter.dev.bdAddr);
				else continue;
				useFilter = 1;
				self->status = BLT_DC_PUSH;
			}
		}
		else if (rb == -1) {
			self->status = BLT_DC_DISCONNECTED;
		}
	/* end main "get" loop */
	} while (self->status != BLT_DC_DISCONNECTED && self->status != BLT_DC_PUSH);
	
	/* --- pushing --- */
	if (self->status == BLT_DC_PUSH) {
		/* init push */
		self->isPush = 1;
		blt_push_init(&self->myPush);
		event = NULL;
		
		#ifdef DEBUG2
		fprintf(stderr, "dc[%d] entering push mode\n", self->myListIndex);
		#endif
		
		do { /* push loop */
		
			if (event == NULL) {
				/* go to sleep here */
				pthread_cond_wait(&self->myPush.sig, &self->myPush.lock);
				blt_push_get(&self->myPush, &event);
				pthread_mutex_unlock(&self->myPush.lock);
				/* wakeup here */

				#ifdef DEBUG3
				fprintf(stderr, "dc[%d] received event signal\n", self->myListIndex);
				#endif

				/* clear read buffer - check for disconnect */
				do {
					rb = readbuf(&buf, self->mySocket, buffer, 1024, 0);
				} while (rb == 1);
				
				if (rb == -1) self->status = BLT_DC_DISCONNECTED;
				else continue;
			}
			else {
				if (event->tracker != NULL) {
				
					#ifdef DEBUG3
					fprintf(stderr, "dc[%d] pushing!\n", self->myListIndex);
					#endif
				
					pthread_rwlock_rdlock(&event->tracker->pubDevicesLock);
					
					dev = event->tracker->pubDevices;
					for (i = 0; i < event->tracker->numPubDevices; i++) {
						/* error? */
						if (dev == NULL) {
							self->status = BLT_DC_DISCONNECTED;
							break;
						}
						/* check filter */
						if (useFilter) {
							if (filter.la != 0 && filter.lg != 0) {
								if (filter.la != dev[i].la || filter.lg != dev[i].lg) {
									continue;
								}
							}
							else if (memcmp(dev[i].dev.bdAddr, filter.dev.bdAddr, 6) != 0) {
								continue;
							}
						}
						if (!blt_dc_printDev(&dev[i], self->mySocket)) {
							self->status = BLT_DC_DISCONNECTED;
							break;
						}
					}
					
					pthread_rwlock_unlock(&event->tracker->pubDevicesLock);
				}
				if (event != NULL) {
					free(event);
					event = NULL;
				}
			
				if (self->status != BLT_DC_DISCONNECTED) {
					pthread_mutex_lock(&self->myPush.lock);
					blt_push_get(&self->myPush, &event);
					pthread_mutex_unlock(&self->myPush.lock);
				}
			}
			
		} while(self->status != BLT_DC_DISCONNECTED);
	}
	
	/* self destroy thread */
	pthread_mutex_lock(&self->server->datas_sem);
	
	#ifdef DEBUG
	fprintf(stderr, "dc[%d] closing connection\n", self->myListIndex);
	#endif
	
	/* destroy readbuf */
	readbufDestroy(&buf);
	
	self->myListIndex = -1;
	close(self->mySocket);
	self->server->numDatas--;
	blt_push_destroy(&self->myPush);
	
	pthread_mutex_unlock(&self->server->datas_sem);
	pthread_exit((void*)NULL);
}

/**
 *  @brief init data client session
 *
 *  @param dcs    pointer to session to init
 *  @param server pointer to server session
 *
 *  @returns nothing
 */
void blt_dc_init(struct blt_data_client_session_t *dcs, struct blt_server_session_t *server)
{
	assert(dcs != NULL);
	assert(server != NULL);

	dcs->isPush = 0;
	dcs->status = BLT_DC_RUNNING;
}
