/*-------------------------------------------------------------------
 *
 *  @file server.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include "blt_server.h"

/**
 *  @brief listen for incoming tracker client connections
 *
 *  This function runs as a separate thread, it listens on the
 *  given ip and port for tracker clients and spawns a new thread 
 *  for each incoming connection.
 *
 *  @param args pointer to the current server session
 *
 *  @returns nothing
 */
void *trackers_wait_for_clients(void *args)
{
	struct blt_server_session_t *server = (struct blt_server_session_t*) args;
	struct blt_tracker_client_session_t *tmpClient;
	int i;
	int rs, ls;
	struct sockaddr_in laddr, raddr;
	struct hostent *hostinfo;
	socklen_t socklen;


	/* prepare socket for binding and listening */
	laddr.sin_port = htons((u_short) server->tracker_port);
	if ((hostinfo = gethostbyname(server->tracker_ip))) {
		bcopy(hostinfo->h_addr, (char *) &laddr.sin_addr, hostinfo->h_length);
		laddr.sin_family = AF_INET;
	}
	else {
		fprintf(stderr, "ts: can't gethostbyname for: %s:%d\n", server->tracker_ip, server->tracker_port);
		return(0);
	}

	if ((ls = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr, "ts: can't create socket\n");
		return(0);
	}
	while (bind(ls, (struct sockaddr*) &laddr, sizeof(struct sockaddr_in)) == -1) {
		#ifdef DEBUG3
		fprintf(stderr, "ts: can't bind to: %s:%d\n", server->tracker_ip, server->tracker_port);
		#endif
		sleep(5);
	}
	if (listen(ls, 2) == -1) {
		fprintf(stderr, "ts: can't listen to socket\n");
		return(0);
	} 
	
	/* accept incoming connections */
	do {
		#ifdef DEBUG2
		fprintf(stderr, "ts: accepting tracker connections at: %s:%d\n", server->tracker_ip, server->tracker_port);
		#endif
		
		if ((rs = accept(ls, (struct sockaddr*) &raddr, &socklen)) != -1) {

			#ifdef DEBUG2
			fprintf(stderr, "ts: new connection\n");
			#endif

			if (server->numTrackers < MAX_NUM_TRACKERS) {
			
				/* lock down the trackers data */
				pthread_mutex_lock(&server->trackers_sem);

				server->numTrackers++;
				
				/* search for a free bucket */
				for (i = 0; i < MAX_NUM_TRACKERS; i++) {
					if (server->trackers[i].myListIndex == -1) {
					
						#ifdef DEBUG2
						fprintf(stderr, "ts: using bucket %d for connection %d\n", i, server->numTrackers);
						#endif
						
						/* setup client environment */
						tmpClient = &server->trackers[i];
						tmpClient->myListIndex = i;
						tmpClient->mySocket = rs;
						tmpClient->server = server;
						memcpy(&tmpClient->clientAl, &socklen, sizeof(socklen_t));
						memcpy(&tmpClient->clientAddr, &raddr, sizeof(raddr));
						
						blt_tc_init(tmpClient, server);
						
						#ifdef BLT_SSL
						tmpClient->ssl = SSL_new(server->ssl_ctx);
						SSL_set_fd(tmpClient->ssl, tmpClient->mySocket);
						#else
						tmpClient->ssl = NULL;
						#endif			
						
						break;
					}
				}
				
				#ifdef DEBUG2
				fprintf(stderr, "ts: before SSL\n");
				#endif
				if (SSL_accept(tmpClient->ssl) == 1) {
					/* start new thread */
					pthread_create(&server->trackers[tmpClient->myListIndex].myThread, NULL, blt_tracker_client,
						(void*) &server->trackers[tmpClient->myListIndex]);
					#ifdef DEBUG2
					fprintf(stderr, "ts: SSL success\n");
					#endif
				}
				else {
					#ifdef DEBUG2
					fprintf(stderr, "ts: SSL failed\n");
					#endif
					SSL_free(tmpClient->ssl);
					server->trackers[i].myListIndex = -1;
					server->numTrackers--;
					close(tmpClient->mySocket);
				}
				
				/* release the trackers data */
				pthread_mutex_unlock(&server->trackers_sem);
			}
			else {
				nwrite(rs, "blt_server: sorry, too many connections, bye!\n", 46);
				sleep(1);
				close(rs);
				
				fprintf(stderr, "ts: connection limit reached, disallowing client connect!\n");
			}
		}
		else {
			fprintf(stderr, "ts: error while accepting new connection!\n");
		}
 	} while (1);
}

/**
 *  @brief initialize the tracker server
 *
 *  @param server pointer to server to init
 *
 *  @returns nothing
 */
void trackers_init(struct blt_server_session_t *server)
{
	int i;
	

	assert(server != NULL);
	server->trackers = (struct blt_tracker_client_session_t*)
		malloc(sizeof(struct blt_tracker_client_session_t) * MAX_NUM_TRACKERS);
	assert(server->trackers != NULL);
	memset(server->trackers, 0, sizeof(struct blt_tracker_client_session_t) * MAX_NUM_TRACKERS);
		
	for (i = 0; i < MAX_NUM_TRACKERS; i++) {
		server->trackers[i].myListIndex = -1;
		
		pthread_mutex_init(&server->trackers[i].actionLock, NULL);
		pthread_cond_init(&server->trackers[i].actionSig, NULL);
		pthread_rwlock_init(&server->trackers[i].pubDevicesLock, NULL);
	}
	pthread_mutex_init(&server->trackers_sem, NULL);
}

/**
 *  @brief listen for incoming data client connections
 *
 *  This function runs as a separate thread, it listens on the
 *  given ip and port for data clients and spawns a new thread 
 *  for each incoming connection.
 *
 *  @param args pointer to the current server session
 *
 *  @returns nothing
 */
void *datas_wait_for_clients(void *args)
{
	struct blt_server_session_t *server = (struct blt_server_session_t*) args;
	struct blt_data_client_session_t *tmpClient;
	int i;
	int rs, ls;
	struct sockaddr_in laddr, raddr;
	struct hostent *hostinfo;
	socklen_t socklen;


	/* prepare socket for binding and listening */
	laddr.sin_port = htons((u_short) server->data_port);
	if ((hostinfo = gethostbyname(server->data_ip))) {
		bcopy(hostinfo->h_addr, (char *) &laddr.sin_addr, hostinfo->h_length);
		laddr.sin_family = AF_INET;
	}
	else {
		fprintf(stderr, "ds: can't gethostbyname for: %s:%d\n", server->data_ip, server->data_port);
		return(0);
	}

	if ((ls = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr, "ds: can't create socket\n");
		return(0);
	} 	
	while (bind(ls, (struct sockaddr*) &laddr, sizeof(struct sockaddr_in)) == -1) {
		#ifdef DEBUG3
		fprintf(stderr, "ds: can't bind to: %s:%d\n", server->data_ip, server->data_port);
		#endif
		sleep(5);
	}
	if (listen(ls, 2) == -1) {
		fprintf(stderr, "ds: can't listen to socket\n");
		return(0);
	} 
	
	/* accept incoming connections */
	do {
		#ifdef DEBUG2
		fprintf(stderr, "ds: accepting data connections at: %s:%d\n", server->data_ip, server->data_port);
		#endif
		
		if ((rs = accept(ls, (struct sockaddr*) &raddr, &socklen)) != -1) {

			#ifdef DEBUG2
			fprintf(stderr, "ds: new connection\n");
			#endif

			if (server->numDatas < MAX_NUM_DATAS) {
			
				/* lock down the datas data */
				pthread_mutex_lock(&server->datas_sem);

				server->numDatas++;
				
				/* search for a free bucket */
				for (i = 0; i < MAX_NUM_DATAS; i++) {
					if (server->datas[i].myListIndex == -1) {
					
						#ifdef DEBUG2
						fprintf(stderr, "ds: using bucket %d for connection %d\n", i, server->numDatas);
						#endif
						
						/* setup client environment */
						tmpClient = &server->datas[i];
						tmpClient->myListIndex = i;
						tmpClient->mySocket = rs;
						tmpClient->server = server;
						memcpy(&tmpClient->clientAl, &socklen, sizeof(socklen_t));
						memcpy(&tmpClient->clientAddr, &raddr, sizeof(raddr));
						
						blt_dc_init(tmpClient, server);
						break;
					}
				}
				
				/* start new thread */
				pthread_create(&server->datas[tmpClient->myListIndex].myThread, NULL, blt_data_client,
					(void*) &server->datas[tmpClient->myListIndex]);
			
				/* release the trackers data */
				pthread_mutex_unlock(&server->datas_sem);
			}
			else {
				nwrite(rs, "blt_server: sorry, too many connections, bye!\n", 46);
				sleep(1);
				close(rs);
				
				fprintf(stderr, "ds: connection limit reached, disallowing client connect!\n");
			}
		}
		else {
			fprintf(stderr, "ds: error while accepting new connection!\n");
		}
 	} while (1);
}

/**
 *  @brief initialize the data server
 *
 *  @param server pointer to server to init
 *
 *  @returns nothing
 */
void datas_init(struct blt_server_session_t *server)
{
	int i;
	

	assert(server != NULL);
	server->datas = (struct blt_data_client_session_t*)
		malloc(sizeof(struct blt_data_client_session_t) * MAX_NUM_DATAS);
	assert(server->datas != NULL);
		
	for (i = 0; i < MAX_NUM_DATAS; i++) {
		server->datas[i].myListIndex = -1;
	}
	pthread_mutex_init(&server->datas_sem, NULL);
}
