/*-------------------------------------------------------------------
 *
 *  @file tracker.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include "blt_server.h"

/**
 *  @brief ask client to authenticate
 *
 *  @param tcs		pointer to current tracker client session
 *
 *  @returns 1 on success 0 on error (connection closed)
 */
int blt_tc_askAuthentication(struct blt_tracker_client_session_t *tcs)
{
	char buffer[1024];
	
	
	assert(tcs != NULL);
	
	
	tcs->msgNo++;
	sprintf(buffer, "req(authentication, %d)\r\n", tcs->msgNo);

	if (tcs->ssl != NULL) {
		if (SSL_write(tcs->ssl, buffer, strlen(buffer)) <= 0) return(0);
		else return(1);
	}
	else {
		if (nwrite(tcs->mySocket, buffer, strlen(buffer)) == -1) return(0);
		else return(1);
	}
}

/**
 *  @brief authenticate the client
 *
 *  @param tcs		pointer to current tracker client session
 *  @param buffer	pointer to command line 
 *
 *  @returns 1 on success (client is authenticated) 0 on failure
 */
int blt_tc_doAuthentication(struct blt_tracker_client_session_t *tcs, char *buffer)
{
	char login[20];
	char password[20];
	struct blt_location_name_t tmp_loc;
	char *p;
	

	assert(tcs != NULL);
	
	memset(login, 0, 20);
	memset(password, 0, 20);
	memset(&tmp_loc, 0, sizeof(struct blt_location_name_t));
	
	p = strstr(buffer, "\"");
	if (p == NULL) return(0);
	scanWithEsc(p+1, login, 20);
	p = strstr(p+strlen(login)+2, "\"");
	if (p == NULL) return(0);
	scanWithEsc(p+1, password, 20);
	
	#ifdef DEBUG5
	fprintf(stderr, "tc[%d] authentication login: \"%s\" password: \"%s\"\n",
		tcs->myListIndex, login, password);
	#endif
		
	if (blt_ln_getlocation(&tcs->server->locationNames, login, &tmp_loc) == 1) {
		if (memcmp(tmp_loc.password, password, strlen(tmp_loc.password)) == 0) {
			memcpy(&tcs->loc, &tmp_loc, sizeof(struct blt_location_name_t));
			
			#ifdef DEBUG
			fprintf(stderr, "tc[%d] is authenticated!\n", tcs->myListIndex);
			#endif
			
			return(1);
		}
	}
	
	#ifdef DEBUG
	fprintf(stderr, "tc[%d] authenticated failed - closing connection!\n", tcs->myListIndex);
	#endif
			
	return(0);
}

/**
 *  @brief fire inquiry request
 *
 *  @param tcs pointer to current tracker client session
 *
 *  @returns 1 on success and 0 on failure
 */
int blt_tc_doInq(struct blt_tracker_client_session_t *tcs)
{
	char buffer[1024];


	assert(tcs != NULL);
	
	tcs->msgNo++;
	sprintf(buffer, "req(inquiry, %d, %d, 0x%.2X)\r\n", tcs->msgNo,
		tcs->cfg->maxNumDevices, tcs->cfg->inquiryDuration);

	if (tcs->ssl != NULL) {
		if (SSL_write(tcs->ssl, buffer, strlen(buffer)) <= 0) return(0);
		else return(1);
	}
	else {
		if (nwrite(tcs->mySocket, buffer, strlen(buffer)) == -1) return(0);
		else return(1);
	}
}

/**
 *  @brief parse a name_request result
 *
 *  @param tcs    pointer to current tracker client session
 *  @param buffer input string to parse
 *
 *  @returns 1 if successfully parsed or 0 on error
 */
int blt_tc_parseAskNameResult(struct blt_tracker_client_session_t *tcs, char *buffer)
{
	char *p;
	char *e;
	struct blt_device_t dev;
	int l;

	
	assert(tcs != NULL);
	
	if (buffer == NULL) return(0);
	l = strlen(buffer);
	if (l < 20) return(0);
	e = buffer + l;
	memset(&dev, 0, sizeof(struct blt_device_t));

	/* ignore "name_request," */
	p = strstr(buffer, ",");
	if (p <= 0) return(0);
	p++;
	if (p > e) return(0);
	
	/* ignore message number */
	p = strstr(p, ",");
	if (p <= 0) return(0);
	p++;
	if (p > e) return(0);
	
	/* ignore spaces/tabs */
	while (*p == ' ' || *p == '\t') {
		p++;
		if (p > e) return(0);
	}
	
	/* get bd_addr */
	parsemac(p, dev.bdAddr);
	
	/* get name */
	p = strstr(p, "\"");
	if (p > 0) scanWithEsc(p+1, dev.name, BLT_DEVICE_NAME_LENGTH+1);
	else return(0);
	
	#ifdef DEBUG
	fprintf(stderr, "tc[%d] name: BD_ADDR: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X Name: %s\n",
		tcs->myListIndex, dev.bdAddr[0],dev.bdAddr[1],dev.bdAddr[2], dev.bdAddr[3],
		dev.bdAddr[4], dev.bdAddr[5], dev.name);
	#endif
	
	/* only save name - if name is given */
	if (strlen(dev.name) > 0) {
		/* save device name */
		blt_dn_write(&tcs->server->devNames, &dev);
	}
	
	return(1);
}

/**
 *  @brief process a result line from an inquiry
 *
 *  @param tcs    pointer to current tracker client session
 *  @param buffer pointer to input buffer
 *
 *  @returns 1 on success or 0 on failure
 */
int blt_tc_parseInqResult(struct blt_tracker_client_session_t *tcs, char *buffer)
{
	struct blt_device_status_t nDev;
	struct tm tmp_time;
	char *p;
	struct blt_device_status_t *tmp;
	

	assert(tcs != NULL);
	assert(buffer != NULL);

	memset(&nDev, 0, sizeof(struct blt_device_status_t));

	p = strstr(buffer, ", ");
	if (p > 0) p += 2;
	else return(0);
	p = strstr(p, ", ");
	if (p > 0) parsemac(p+2, nDev.dev.bdAddr);
	else return(0);
	p = strstr(p, "0x");
	if (p > 0) nDev.dev.devClass = strtoul(p, NULL, 16);
	else return(0);
	p += 2;
	p = strstr(p, "0x");
	if (p > 0) nDev.rssi = strtoul(p, NULL, 16);
	else return(0);
	p = strstr(p, "\"");
	memset(&tmp_time, 0, sizeof(struct tm));
	sscanf(p, "\"%d.%d.%d %d:%d:%d\", %d, %d)", &tmp_time.tm_mon,
		&tmp_time.tm_mday,	&tmp_time.tm_year, &tmp_time.tm_hour, &tmp_time.tm_min,
		&tmp_time.tm_sec, &nDev.la, &nDev.lg);
		
	tmp_time.tm_year -= 1900;
	tmp_time.tm_mon--;
	tmp_time.tm_hour--;
	nDev.tStamp = mktime(&tmp_time);
	
	#ifdef DEBUG
	{
	char ts[30];
	memset(&tmp_time, 0, sizeof(struct tm));
	localtime_r(&nDev.tStamp, &tmp_time);
	strftime(ts, 21, "%m.%d.%Y %T", &tmp_time);
	fprintf(stderr, "tc[%d] inquiry: BD_ADDR: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X, CLASS: 0x%.6X, RSSI: 0x%.4X, Latitude: %d, Longitude: %d, Time: %s\n",
		tcs->myListIndex, nDev.dev.bdAddr[0], nDev.dev.bdAddr[1], nDev.dev.bdAddr[2], nDev.dev.bdAddr[3],
		nDev.dev.bdAddr[4], nDev.dev.bdAddr[5],nDev.dev.devClass, nDev.rssi, nDev.la, nDev.lg, ts);
	}
	#endif

	if (tcs->maxPrvDevices == 0) {
		tcs->maxPrvDevices = 50;
		tcs->prvDevices = malloc(tcs->maxPrvDevices * sizeof(struct blt_device_status_t));
		memset(tcs->prvDevices, 0, tcs->maxPrvDevices * sizeof(struct blt_device_status_t));
	}
	else if (tcs->numPrvDevices == tcs->maxPrvDevices) {
		tmp = malloc((tcs->maxPrvDevices + (tcs->maxPrvDevices / 2)) * sizeof(struct blt_device_status_t));
		memset(tmp, 0, (tcs->maxPrvDevices + (tcs->maxPrvDevices / 2)) * sizeof(struct blt_device_status_t));
		memcpy(tmp, tcs->prvDevices, tcs->numPrvDevices * sizeof(struct blt_device_status_t));
		free(tcs->prvDevices);
		tcs->prvDevices = tmp;
		tcs->maxPrvDevices = tcs->maxPrvDevices + (tcs->maxPrvDevices / 2);
	}
	
	/* try to fill in the device name */
	blt_dn_read(&tcs->server->devNames, &nDev.dev);
	/* copy device status struct */
	memcpy(&tcs->prvDevices[tcs->numPrvDevices], &nDev, sizeof(struct blt_device_status_t));
	
	/* increment device count */
	tcs->numPrvDevices++;
	
	return(1);
}

/**
 *  @brief move private data to public data
 *
 *  @param tcs pointer to current tracker client session
 *
 *  @returns nothing
 */
void blt_tc_prvToPub(struct blt_tracker_client_session_t *tcs)
{
	assert(tcs != NULL);

	#ifdef DEBUG5
	fprintf(stderr, "prvToPub: in\n");
	#endif
	
	pthread_rwlock_wrlock(&tcs->pubDevicesLock);
	
	#ifdef DEBUG5
	fprintf(stderr, "prvTpPub1: %d\n", (unsigned int) tcs->pubDevices);
	#endif
	
	if (tcs->pubDevices != NULL) free(tcs->pubDevices);
	
	#ifdef DEBUG5
	fprintf(stderr, "prvTpPub2: %d\n", tcs->numPrvDevices);
	#endif
	
	if (tcs->numPrvDevices == 0) tcs->pubDevices = NULL;
	else  tcs->pubDevices = tcs->prvDevices;
	tcs->numPubDevices = tcs->numPrvDevices;
	tcs->numPrvDevices = 0;
	tcs->maxPrvDevices = 0;
	
	pthread_rwlock_unlock(&tcs->pubDevicesLock);
	
	#ifdef DEBUG5
	fprintf(stderr, "prvToPub: out\n");
	#endif
}

/**
 *  @brief send configuration to client
 *
 *  @param tcs pointer to current tracker client session
 *  @param inq != 0 indicates that the special inquiry config. should be used
 *
 *  @returns 1 on success and 0 on failure
 */
int blt_tc_conf(struct blt_tracker_client_session_t *tcs, int inq)
{
	char buffer[1024];
	char name[256];

	
	assert(tcs != NULL);
	
	/*  build device name
	 *
	 *  Looks like "BLT_#1"
	 */
	if (strlen(tcs->loc.devname) == 0) {
		sprintf(name, BLT_TRACKER_NAME, tcs->myListIndex+1);
	}
	else {
		strcpy(name, tcs->loc.devname);
	}

	tcs->msgNo++;
	
	/* normal configuration */
	if (inq == 0) {
		sprintf(buffer, "req(configuration, %d, %d, %d, 0x%.4X, \"%s\")\r\n",
			tcs->msgNo, tcs->cfg->pageScan, tcs->cfg->inquiryScan,
			tcs->cfg->pageTimeout, name);
	}
	/* special "before" inquiry configuration */
	else {
		sprintf(buffer, "req(configuration, %d, 0, 0, 0x%.4X, \"\")\r\n",
			tcs->msgNo, tcs->cfg->pageTimeout);
	}
	
	if (tcs->ssl != NULL) {
		if (SSL_write(tcs->ssl, buffer, strlen(buffer)) <= 0) return(0);
		else return(1);
	}
	else {
		if (nwrite(tcs->mySocket, buffer, strlen(buffer)) == -1) return(0);
		else return(1);
	}
}

/**
 *  @brief send out name_requests for unknown devices
 *
 *  @param tcs pointer to current tracker client session
 *
 *  @returns 1 for success or 0 for failure
 */
int blt_tc_askName(struct blt_tracker_client_session_t *tcs)
{
	int i;
	struct blt_device_t *dev;
	char buffer[1024];
	
	
	assert(tcs != NULL);
	
	if (tcs->askNameNum >= tcs->cfg->maxAskName) return(0);
	
	for (i = tcs->askNameCur; i < tcs->numPubDevices; i++) {
		if (blt_dn_read(&tcs->server->devNames, &tcs->pubDevices[i].dev) == 0) {
			tcs->askNameCur = i + 1;
			tcs->askNameNum++;	
			dev = &tcs->pubDevices[i].dev;
			tcs->msgNo++;
			sprintf(buffer, "req(name_request, %d, %.2X:%.2X:%.2X:%.2X:%.2X:%.2X)\r\n", tcs->msgNo,
				dev->bdAddr[0],dev->bdAddr[1],dev->bdAddr[2],dev->bdAddr[3],dev->bdAddr[4],dev->bdAddr[5]);
			
			/*
			 *  all ways return 1 (success) even if write fails
			 *  on success read will be called, read then detects the disconnect
			 */
			/* nwrite(tcs->mySocket, buffer, strlen(buffer)); */
			
			if (tcs->ssl != NULL) SSL_write(tcs->ssl, buffer, strlen(buffer));
			else nwrite(tcs->mySocket, buffer, strlen(buffer));
			return(1);
		}
	}
	return(0);
}

/**
 *  @brief this is a tracker client thread
 *
 *  @params args pointer to current tracker client session
 *
 *  @returns nothing
 */
void *blt_tracker_client(void *args)
{
	struct blt_tracker_client_session_t *self = (struct blt_tracker_client_session_t*) args; /**< self reference */
	char buffer[1024];	/**< receive buffer */
	int rb = 0;				/**< received bytes */
	int wait = 5;			/**< wait for read */
	struct blt_push_t *pushEvent = NULL; /**< inquiry complete event for push clients */
	struct readbuf_t buf;


	assert(args != NULL);
	
	/* init readbuf */
	readbufInit(&buf, 4096, self->ssl);
	
	#ifdef DEBUG
	fprintf(stderr, "tc[%d] ready - not authenticated jet!\n", self->myListIndex);
	#endif

	/* --- ask to authenticat --- */
	blt_tc_askAuthentication(self);

	do {	/* authentication loop */
		/* READ */
		memset(buffer, 0, 1024);
		if ((rb = readbuf(&buf, self->mySocket, buffer, 1024, 10)) == 1) {
		
			if (strstr(buffer, "ind(authentication, ") == buffer) {
				if (blt_tc_doAuthentication(self, buffer) == 1) break;
				else self->status = BLT_TC_DISCONNECTED;
			}
			else {
				self->status = BLT_TC_DISCONNECTED;
			}
		}
		else {
			self->status = BLT_TC_DISCONNECTED;
		}
	} while (self->status != BLT_TC_DISCONNECTED); /* authentication loop */
	
	
	if (self->status == BLT_TC_DISCONNECTED) {
			/*
			 * I know, this is soooo evil!
			 */
			goto disconnected;
			/*
			 * jump to cleanup code at end of function
			 */
	}

	#ifdef DEBUG
	fprintf(stderr, "tc[%d] ready!\n", self->myListIndex);
	#endif

	/* --- config client --- */
	blt_tc_conf(self, 0);
		
	do {	/* main loop */

		/* inquiry notification */	
		if (self->doInq) {
			self->status = BLT_TC_DOINQ;
		}
	
		/* SLEEP */
		if (self->status == BLT_TC_SLEEP) {
			self->status = BLT_TC_SLEEPING;
			/* block here */
			pthread_mutex_lock(&self->actionLock);
			pthread_cond_wait(&self->actionSig, &self->actionLock);
			pthread_mutex_unlock(&self->actionLock);
			/* continue here on wakeup */
			wait = 5;
			continue;
		}
		/* INQUIRY */
		else if (self->status == BLT_TC_DOINQ) {
			#ifdef DEBUG2
			fprintf(stderr, "tc[%d] starting inquiry\n", self->myListIndex);
			#endif
		
			/* tell client to switch of scanning */
			if (!blt_tc_conf(self, 1)) {
				/* error - disconnect */
				self->status = BLT_TC_DISCONNECTED;
				break;
			}		
			
			/* begin inquiry */
			if (!blt_tc_doInq(self)) {
				/* error - disconnect */
				self->status = BLT_TC_DISCONNECTED;
				break;
			}
			
			/* reset inquiry flag */
			self->doInq = 0;
			/* switch to read mode */
			self->status = BLT_TC_READ;
			
			/* calc. new wait time */
			wait = (self->cfg->inquiryDuration * BLUETOOTH_SLOT) + 1;
			
			#ifdef DEBUG2
			fprintf(stderr, "tc[%d] inquiry result wait: %d\n",self->myListIndex, wait);
			#endif
		}
		/* ASK FOR NAMES */
		else if (self->status == BLT_TC_ASKNAME) {
			if (!blt_tc_askName(self)) self->status = BLT_TC_SLEEP;
			/* calc. new wait time */
			wait = (self->cfg->pageTimeout * BLUETOOTH_SLOT) + 1;
		}
		
		/* READ */
		memset(buffer, 0, 1024);
		if ((rb = readbuf(&buf, self->mySocket, buffer, 1024, wait)) == 1) {
		
			#ifdef DEBUG2
			fprintf(stderr, "line: %s\n", buffer);
			#endif
			
			if (strstr(buffer, "ind(inquiry_result_event") == buffer) {
				blt_tc_parseInqResult(self, buffer);
			}
			else if (strstr(buffer, "ind(inquiry_complete") == buffer) {
				#ifdef DEBUG2
				fprintf(stderr, "tc[%d] inquiry_complete\n", self->myListIndex);
				#endif

				/*  tell client to switch scanning back on
				 *  - this is needed for request_name (askName)
				 */
				if (!blt_tc_conf(self, 0)) {
					/* error - disconnect */
					self->status = BLT_TC_DISCONNECTED;
					break;
				}
				
				#ifdef DEBUG5
				fprintf(stderr, "tc[%d] DEBUG before prvToPub\n", self->myListIndex);
				#endif
					
				/* copy results to public memory */
				blt_tc_prvToPub(self);
				
				#ifdef DEBUG5
				fprintf(stderr, "tc[%d] DEBUG after prvToPub\n", self->myListIndex);
				#endif
				
				if (self->numPubDevices > 0) {
					/* notify pushServer */
					pushEvent = malloc(sizeof(struct blt_push_t));
					assert(pushEvent != NULL);
					/* pushEvent->last = NULL;
					pushEvent->next = NULL; */
					pushEvent->tracker = self;
					blt_push_put(&self->server->pushServer, pushEvent);
					pthread_cond_signal(&self->server->pushServer.sig);
					pushEvent = NULL;
					
					#ifdef DEBUG5
					fprintf(stderr, "tc[%d] DEBUG after event push\n", self->myListIndex);
					#endif
				}
					
				/* switch to askname */
				self->status = BLT_TC_ASKNAME;
				self->askNameCur = 0;
				self->askNameNum = 0;
			}
			else if (strstr(buffer, "ind(name_request") == buffer) {
				blt_tc_parseAskNameResult(self, buffer);
			}
		}
		else if (rb == -1) {
			self->status = BLT_TC_DISCONNECTED;
		}
		else {
			if (self->status != BLT_TC_ASKNAME) self->status = BLT_TC_SLEEP;
		}
		
	} while (self->status != BLT_TC_DISCONNECTED); /* main loop */

/*
 *  evil jump point
 */
disconnected:

	/* self destroy thread */
	pthread_mutex_lock(&self->server->trackers_sem);

	readbufDestroy(&buf);
	
	#ifdef DEBUG
	fprintf(stderr, "tc[%d] closing connection\n", self->myListIndex);
	#endif
	
	close(self->mySocket);
	self->myListIndex = -1;
	self->server->numTrackers--;
	
	#ifdef BLT_SSL
	SSL_free(self->ssl);
	#endif

	pthread_mutex_unlock(&self->server->trackers_sem);
	pthread_exit((void*)NULL);
}

/**
 *	 @brief initialize a tracker client before starting the thread
 *
 *  @param tcs    pointer to current tracker client session
 *  @param server pointer to current server session
 *
 *  @returns nothing
 */
void blt_tc_init(struct blt_tracker_client_session_t *tcs, struct blt_server_session_t *server)
{
	assert(tcs != NULL);
	assert(server != NULL);

	tcs->numPrvDevices = 0;
	tcs->maxPrvDevices = 0;
	/* tcs->prvDevices = NULL; */
	tcs->numPubDevices = 0;
	/* tcs->pubDevices = NULL; */

	tcs->msgNo = 0;
	tcs->status = BLT_TC_SLEEP;
	tcs->doInq = 0;
	tcs->cfg = &server->defaultCfg;
	
	memset(&tcs->loc, 0, sizeof(struct blt_location_name_t));
}
